"""Privacy metrics."""

import numpy as np
from sklearn.neighbors import LocalOutlierFactor, NearestNeighbors

from .. import utils


def _get_weap_scores(synth, key, target):
    """Within Equivalence class Attribution Probabilities (WEAP).

    For each record in the synthetic dataset, this function returns the
    proportion across the whole dataset that a set of `key` values are
    matched with this `target` value.

    Parameters
    ----------
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    key : list of str
        List of features in `synth` to use as the key.
    target : str
        Feature to use as the target.

    Returns
    -------
    pandas.Series
        A series object containing the WEAP scores for each record in
        `synth`.

    Notes
    -----
    This function is intended to only be used within `TCAP()` to
    determine which synthetic records are most likely to pose an
    attribution risk.
    """

    synth = synth.copy()
    key_and_target = [*key, target]

    key_target_vc = synth.value_counts(subset=key_and_target).reset_index()
    key_target_vc.columns = key_and_target + ["key_target_frequency"]
    key_vc = synth.value_counts(key).reset_index()
    key_vc.columns = key + ["key_frequency"]

    synth = synth.merge(key_target_vc)
    synth = synth.merge(key_vc)

    return synth["key_target_frequency"] / synth["key_frequency"]


def tcap_score(real, synth, key, target):
    """Target Correct Attribution Probability (TCAP) score.

    This privacy metric calculates the average chance that the
    key-target pairings in a synthetic dataset reveal the true
    key-target pairings in associated real dataset.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    key : list of str
        List of features in `synth` to use as the key.
    target : str
        Feature to use as the target.

    Returns
    -------
    float
        The average TCAP score across the dataset.

    Notes
    -----
    This metric provides an estimate of how well an intruder could infer
    attributes of groups in the real dataset by studying the synthetic.
    The choices for `key` and `target` will vary depending on the
    dataset in question but we would suggest the `key` features are
    those that could be readily available to an outsider and the
    `target` feature is one we wouldn't want them finding out, such as a
    protected characteristic.

    This method only works with categorical data, so binning of
    continuous data may be required.

    Full details may be found in:

    Taub and Elliott (2019). The Synthetic Data Challenge. The Hague,
    The Netherlands: Joint UNECE/Eurostat Work Session on Statistical
    Data Confidentiality, Session 3.
    """

    scores = _get_weap_scores(synth, key, target)

    if sum(scores == 1) == 0:
        return 0

    synth_reduced = synth[scores == 1][[*key, target]]
    synth_reduced.drop_duplicates(inplace=True)
    synth_reduced.rename(
        columns={target: target + "_synthguess"}, inplace=True
    )

    combined = real.merge(synth_reduced, how="left", on=key)
    target_matches = combined[target] == combined[target + "_synthguess"]

    return np.mean(target_matches)


def _find_outliers(data, threshold, n_neighbours):
    """Identify local outliers using the nearest-neighbour principle.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe to be assessed for outliers.
    threshold : float
        Float influencing classification of outliers. Increasing this
        threshold means that fewer points are considered outliers.
    n_neighbours : int
        Number of neighbours to consider in outlier detection.

    Returns
    -------
    outlier_bool : list of bool
        List indicating which rows of `data` are outliers.

    Notes
    -----
    Most inliers will have an outlier factor of less than one, however
    there are no clear rules that determine when a data point is an
    outlier. This is likely to vary from dataset to dataset and, as
    such, we recommend tuning `outlier_factor_threshold` to suit.
    """

    lof = LocalOutlierFactor(n_neighbors=n_neighbours)
    lof.fit_predict(data)

    outlier_factor = -lof.negative_outlier_factor_
    outlier_bool = outlier_factor < threshold

    return outlier_bool


def min_nearest_neighbour(
    real,
    synth,
    feats=None,
    outliers_only=True,
    threshold=2,
    n_neighbours=5,
):
    """Minimum nearest-neighbour distance.

    This privacy metric returns the smallest distance between any point
    in the real dataset and any point in the synthetic dataset. There is
    an option to only consider the outliers in the real dataset as these
    perhaps pose more of a privacy concern.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feats : list of str or None, default None
        Features in `real` and `synth` to use when calculating
        distance. If `None` (default), all common features are used.
    outliers_only : bool, default True
        Boolean indicating whether to filter out the real data inliers
        (default) or not.
    threshold : number, default 2
        Outlier decision threshold. Increase to include fewer points
        from `real` in nearest-neighbour calculations.
    n_neighbours : int, default 5
        Number of neighbours to consider when identifying local
        outliers.

    Returns
    -------
    float
        Minimum Manhattan distance between `real` and `synth` data.

    Notes
    -----
    This privacy metric provides an insight into whether the synthetic
    dataset is too similar to the real dataset. It does this by
    calculating the minimum distance between the real records and the
    synthetic records.

    This metric assumes that categorical data is ordinal during distance
    calculations, or that it has already been suitably one-hot-encoded.
    """

    combined = utils.df_combine(real, synth, feats=feats)
    combined_recode, _ = utils.cat_encode(combined, return_all=True)
    real, synth = utils.df_separate(
        combined_recode,
        source_col_name="source",
        source_val_real=0,
        source_val_synth=1,
    )

    if outliers_only:
        outliers = _find_outliers(real, threshold, n_neighbours)
        real = real[outliers]

    neigh = NearestNeighbors(n_neighbors=1, radius=100, p=1).fit(real)
    distances, _ = neigh.kneighbors(synth, return_distance=True)

    return np.min(distances)


def _get_sample(data, feats, n_samples, seed, label):
    """Take a sample from the data and count the feature frequencies."""

    return (
        data[feats]
        .sample(n_samples, random_state=seed)
        .assign(**{f"{label}_count": 1})
        .groupby(feats)
        .count()
        .reset_index()
    )


def sample_overlap_score(
    real,
    synth,
    feats=None,
    sample_size=0.2,
    runs=5,
    seed=None,
    score_type="unique",
):
    """Return percentage of overlap between real and synth data based on
    random sampling.

    Samples from both the real and synthetic datasets are compared for
    similarity. This similarity, or overlap score, is based on the
    exact matches of real data records within the synthetic data.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feats : list of str or None, default None
        The features used to match records. If `None` (default), all
        common features are used.
    sample_size : float or int, default 0.2
        The ratio (if `sample_size` between 0 and 1) or count
        (`sample_size` > 1) of records to sample. Default is 0.2 (20%).
    runs : int, default 5
        The number of sampling runs to use when computing the score.
    seed : int, optional
        Random number seed used for sampling.
    score_type : {"unique", "sample"}, default "unique"
        Method used for calculating the overlap score. If "unique"
        (default), the score is the percentage of unique records in the
        real sample that have a match within the synthetic data. If
        "sample", the score is the percentage of all records within the
        real sample that have a match within the synth sample.

    Returns
    -------
    overlap_score : float
        Estimated overlap score between `real` and `synth`.
    """

    feats = feats or real.columns.intersection(synth.columns).to_list()

    min_num_rows = min(real.shape[0], synth.shape[0])
    if 0 <= sample_size <= 1:
        n_samples = int(min_num_rows * sample_size)
    else:
        n_samples = min(min_num_rows, sample_size)

    scores = []
    for _ in range(runs):

        sample_real = _get_sample(real, feats, n_samples, seed, "real")
        sample_synth = _get_sample(synth, feats, n_samples, seed, "synth")

        duplicates = sample_real.merge(
            sample_synth,
            how="left",
            on=feats,
            suffixes=("_real", "_synth"),
            indicator="_match",
        )

        if score_type == "unique":
            score = duplicates._match.value_counts(normalize=True).both
        if score_type == "sample":
            score = (
                duplicates[duplicates._match == "both"].real_count.sum()
                / n_samples
            )

        scores.append(score)

    return np.mean(scores)



"""Correlation-based utility metrics."""

import itertools

import numpy as np
import pandas as pd
from scipy.stats import chi2_contingency


def _mean_squared_difference(x, y):
    """Calculate the mean-squared difference (error) between two numeric
    objects or two arrays."""

    return np.nanmean((np.array(x) - np.array(y)) ** 2)


def _cramers_v(var1, var2):
    """Cramer's V.

    Measures the association between two nominal categorical variables.

    Parameters
    ----------
    var1 : pandas.Series
        Measurements for the first variable.
    var2 : pandas.Series
        Measurements for the second variable.

    Returns
    -------
    float
        The association between the two variables.

    Notes
    -----
    Wikipedia suggests that this formulation of Cramer's V tends to
    overestimate the strength of an association and poses a corrected
    version. However, since we are only concerned with how associations
    compare and not what the actual values are, we continue to use this
    simpler version.
    """

    confusion = np.array(pd.crosstab(var1, var2))
    chi2, *_ = chi2_contingency(confusion, correction=False)

    return np.sqrt((chi2 / np.sum(confusion)) / (min(confusion.shape) - 1))


def _pairwise_cramers_v(data):
    """Compute pairwise Cramer's V for the columns of `data`."""

    results = []
    for x, y in itertools.product(data.columns, repeat=2):
        results.append(_cramers_v(data[x], data[y]))

    size = data.shape[1]
    results = np.array(results).reshape((size, size))

    return pd.DataFrame(results, index=data.columns, columns=data.columns)


def correlation_msd(real, synth, method="pearson", feats=None):
    """Mean-squared difference in correlation coefficients.

    This metric calculates the mean squared difference between the
    correlation matrices for the real and synthetic datasets. This gives
    an indication of how well the synthetic data has retained bivariate
    relationships.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    method : {"pearson", "spearman", "cramers_v"}, default "pearson"
    feats : list of str or None, default None
        Features to measure correlation across. If `method="cramers_v"`,
        all numeric columns will be filtered out. Likewise, for the
        other correlation methods, all non-numeric columns are removed.
        If `None` (default), all common features that satisfy the needs
        of `method` are used.

    Returns
    -------
    float
        Mean-squared difference of correlation coefficients.

    See Also
    --------
    numpy.corrcoef

    Notes
    -----
    The smaller the number this function returns, the better the
    synthetic data captures the correlation between variables in the
    real data. This method is therefore good for comparing multiple
    synthetic datasets to see which is closest to the real. However, as
    this is not a test, there is no threshold value below which we can
    claim the datasets are statistically the same.

    We only consider the coefficients above the main diagonal when
    calculating the MSD. If we included the entire matrix, we would
    double-count each pair of features as well as including the trivial
    ones along the main diagonal.
    """

    feats = feats or real.columns.intersection(synth.columns)

    if method == "cramers_v":
        real = real[feats].select_dtypes(exclude="number")
        synth = synth[feats].select_dtypes(exclude="number")
        rcorr, scorr = _pairwise_cramers_v(real), _pairwise_cramers_v(synth)
    else:
        real = real[feats].select_dtypes(include="number")
        synth = synth[feats].select_dtypes(include="number")
        rcorr, scorr = real.corr(method=method), synth.corr(method=method)

    idxs = np.triu(np.ones(len(rcorr)), k=1).astype(bool)
    rcorr, scorr = rcorr.where(idxs), scorr.where(idxs)

    return _mean_squared_difference(rcorr, scorr)


def _correlation_ratio(categorical, continuous):
    """Categorical-continuous correlation ratio.

    Calculates the correlation ratio for categorical-continuous
    association. Describes the possibility of deducing the corresponding
    category for a given continuous value.

    Missing values are not permitted in either series. Any rows with a
    missing value are dropped from both series before calculating the
    ratio.

    Returns a value in the range [0, 1] where 0 means a category can not
    be determined given a continuous measurement and 1 means it can with
    absolute certainty.

    Parameters
    ----------
    categorical : pandas.Series
        Categorical feature measurements.
    continuous : pandas.Series
        Continuous feature measurements.

    Returns
    -------
    float
        The categorical-continuous association ratio.

    Notes
    -----
    See https://en.wikipedia.org/wiki/Correlation_ratio for details.
    """

    combined = pd.concat((categorical, continuous), axis=1).dropna()
    categorical, continuous = combined.values.T

    categories = np.unique(categorical)
    category_means = np.zeros(len(categories))
    category_counts = np.zeros(len(categories))

    for i, cat in enumerate(categories):
        cts_in_cat = continuous[categorical == cat]
        category_means[i] = np.mean(cts_in_cat)
        category_counts[i] = len(cts_in_cat)

    total_mean = np.mean(continuous)
    numerator = np.sum(category_counts * ((category_means - total_mean) ** 2))
    denominator = np.sum((continuous - total_mean) ** 2)

    return np.sqrt(numerator / denominator)


def correlation_ratio_msd(real, synth, categorical=None, numeric=None):
    """Correlation ratio mean-squared difference.

    This metric calculates the mean-squared difference in association
    between categorical and continuous feature pairings in the real and
    synthetic datasets.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    categorical : list of str or None, default None
        Categorical features in `real` and `synth` to include in
        comparison. If `None` (default), uses all common object-type
        columns.
    numeric : list of str or None, default None
        Numerical features in `real` and `synth` to include in
        comparison. If `None` (default), uses all common columns not
        selected by `categorical`.

    Returns
    -------
    float
        Mean squared difference between `real` and `synth` in
        correlation ratio scores across all categorical-continuous
        feature pairs.
    """

    common = real.columns.intersection(synth.columns)
    categorical = (
        categorical
        or real[common].select_dtypes(include=("object", "category")).columns
    )
    numeric = numeric or common.difference(categorical)

    real_corr_ratio, synth_corr_ratio = [], []
    for cat_feat, num_feat in itertools.product(categorical, numeric):
        real_corr_ratio.append(
            _correlation_ratio(real[cat_feat], real[num_feat])
        )
        synth_corr_ratio.append(
            _correlation_ratio(synth[cat_feat], synth[num_feat])
        )

    return _mean_squared_difference(real_corr_ratio, synth_corr_ratio)



"""Functions for the generic measures from the 2018 NIST competition."""

import itertools

import numpy as np
import pandas as pd


def _numeric_edges(real, synth, bins):
    """Find the bin edges for the numeric features."""

    numeric = real.select_dtypes(include="number").columns.intersection(
        synth.columns
    )
    edges = {col: np.histogram_bin_edges(real[col], bins) for col in numeric}

    return edges


def _discretise_datasets(real, synth, bins):
    """Discretise the numeric features of each dataset."""

    rout, sout = real.copy(), synth.copy()
    edges = _numeric_edges(rout, sout, bins)
    for col, edge in edges.items():
        rout.loc[:, col] = pd.cut(rout[col], edge)
        sout.loc[:, col] = pd.cut(sout[col], edge)

    return rout, sout


def _kway_marginal_score(real, synth, features):
    """Get the transformed score for a single set of features.

    Note that the datasets should have their numeric features
    discretised already.
    """

    rmarg = real.groupby(features).size() / len(real)
    smarg = synth.groupby(features).size() / len(synth)

    return 1 - sum(abs(rmarg - smarg)) / 2


def kway_marginals(real, synth, k=3, trials=100, bins=100, seed=None):
    r"""A measure based on the similarity of a set of k-way marginals.

    This measure works as follows:

        1. Discretise all numeric features (based on the orignal data).
        2. Randomly select `trials` sets of `k` features and calculate
           the corresponding marginal counts for each dataset.
        3. Calculate summed absolute deviation in the counts across all
           bins and marginal sets.
        4. Transform the summed absolute deviations, :math:`s`, to form
           a set of scores :math:`S = \left[1-s/2 | for each s\right]`.
           This transformation ensures the scores are in :math:`[0, 1]`.
           These extremes represent the worst- and best-case scenarios,
           respectively.
        5. Return the mean score.

    The NIST competition utilised a set of 100 three-way marginals.
    Details can be found at https://doi.org/10.6028/NIST.TN.2151.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    k : int, default 3
        Number of features to include in each k-way marginal. Default
        uses 3 (as done in the NIST competition).
    trials : int, default 100
        Maximum number of marginals to consider to estimate the overall
        score. If there are fewer `k`-way combinations than `trials`,
        tries all. Default uses 100 (as done in the NIST competition).
    bins : int or str, default 100
        Binning method for sampled numeric features. Can be anything
        accepted by `numpy.histogram`. Default uses 100 bins (as done in
        the NIST competition).
    seed : int or None, default None
        Random number seed. If `None`, results will not be reproducible.

    Returns
    -------
    score : float
        The mean transformed sum absolute deviation in k-way densities.
    """

    disreal, dissynth = _discretise_datasets(real, synth, bins)
    prng = np.random.default_rng(seed)

    choices = list(
        itertools.combinations(real.columns.intersection(synth.columns), r=k)
    )
    marginals = prng.choice(
        choices, size=min(trials, len(choices)), replace=False
    ).tolist()

    scores = [
        _kway_marginal_score(disreal, dissynth, marginal)
        for marginal in marginals
    ]

    return np.mean(scores)


def _make_rule(data, row, column, prng):
    """Given a column, make a rule for it."""

    values = data[column].unique()
    observed = row[column]

    if pd.api.types.is_numeric_dtype(values):
        rule = (observed, prng.uniform(0, values.max() - values.min()))
    else:
        rule = {observed}
        while True:
            new = prng.choice(values)
            if new in rule:
                break
            rule.add(new)

    return rule


def _create_test_cases(data, trials, prob, seed):
    """Create a collection of HOC test cases.

    For each test case, sample a row. Iterate over the columns,
    including them with some probability and generating them a rule for
    the test case. This rule is determined by the data type of the
    column:

      - Numeric columns use a random subrange from the whole dataset
      - Categoric columns use a random subset of the elements in the
        entire dataset

    Both of these types of rules always include the observed value in
    the row of the associated column; this means that the test will
    always be satisfied by at least one row when it comes to evaluation.
    """

    prng = np.random.default_rng(seed)

    cases = []
    for _ in range(trials):
        row = data.iloc[prng.integers(0, len(data)), :]
        case = {
            column: _make_rule(data, row, column, prng)
            for column in data.columns
            if prng.random() <= prob
        }
        cases.append(case)

    return cases


def _evaluate_test_cases(data, cases):
    """Evaluate the test cases on a dataset.

    Each test case's score is set as the proportion of the dataset for
    which all rules in the test case are satisfied. Each type of rule is
    satisfied differently:

      - Numeric rules are satisfied if the observed value lies within
        the rule's subrange
      - Categoric rules are satisfied if the observed value lies in the
        rule's subset
    """

    results = []
    for case in cases:
        result = pd.DataFrame()
        for col, rule in case.items():
            if isinstance(rule, tuple):
                result[col] = abs(data[col] - rule[0]) <= rule[1]
            else:
                result[col] = data[col].isin(rule)

        results.append(result.min(axis=1).mean())

    return results


def hoc(real, synth, trials=300, prob=0.1, seed=None):
    r"""A measure based on Higher Order Conjunctions (HOC).

    This measure compares the relative sizes of randomly selected pools
    of "similar" rows in the real and synthetic data. This measure of
    similarity is defined across a set of randomly genereated test
    cases applied to each dataset. Each test case consists of a set of
    rules.

    The :math:`i`-th test calculates the fraction of records satisfying
    its rules in the real data, :math:`f_{ri}`, and the synthetic,
    denoted :math:`f_{si}`. Their dissimilarity in test :math:`i` is
    quantified as:

    .. math::

        d_i = \ln\left(\max(f_{si}, 10^{-6})\right) - \ln(f_{ri})

    These dissimilarities are summarised as:

    .. math::

        \Delta = \sqrt{\frac{1}{N} \sum_{i=1}^{N} d_i^2}

    where :math:`N` is the number of test cases. Finally, this is
    transformed to a HOC score:

    .. math::

        HOC = \max \left(0, 1 + \frac{\Delta}{\ln(10^{-3})}\right)

    This measure is bounded between 0 and 1, indicating whether the
    datasets are nothing alike or identical based on the test cases,
    respectively. In the original text this score is multiplied by 1000
    to make it human-readable. Full details are available in
    https://doi.org/10.6028/NIST.TN.2151.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    trials : int, default 300
        Number of test cases to create. Default of 300 as in the
        competition.
    prob : float, default 0.1
        Probability of any column being included in a test case. Default
        of 0.1 as in the competition.
    seed : int or None, default None
        Random number seed. If `None`, results will not be reproducible.

    Returns
    -------
    score : float
        The overall HOC score.

    Notes
    -----
    It is possible that some test cases will be "empty", i.e. when no
    columns are selected. In this scenario, the score for that case will
    be `np.nan` rather than it being resampled.
    """

    cases = _create_test_cases(real, trials, prob, seed)
    real_scores = _evaluate_test_cases(real, cases)
    synth_scores = _evaluate_test_cases(synth, cases)

    dissims = (
        np.log(max(si, 1e-6)) - np.log(ri)
        for ri, si in zip(real_scores, synth_scores)
    )

    delta = np.sqrt(sum(d**2 for d in dissims) / trials)
    score = max(0, 1 + delta / np.log(1e-3))

    return score



"""Propensity-based utility metrics."""

from collections import namedtuple

import numpy as np
import pandas as pd
import scipy.special
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.tree import DecisionTreeClassifier

from ..utils import df_combine


def _combine_encode_and_pop(real, synth):
    """Get the combined, encoded real and synthetic data, and their
    origins.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.

    Returns
    -------
    combined : pandas.DataFrame
        The combined data with categorical columns one-hot encoded.
    indicator : numpy.ndarray
        An indicator for whether the data is real (0) or synthetic (1).
    """

    combined = df_combine(real, synth, source_val_real=0, source_val_synth=1)
    combined = pd.get_dummies(combined, drop_first=True)
    indicator = combined.pop("source").values

    return combined, indicator


def _get_propensity_scores(data, labels, method, **kwargs):
    """Fit a propensity model to the data and extract its scores.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe to fit propensity model to.
    labels : numpy.ndarray
        Indicator for which data are real (0) or synthetic (1).
    method : {"cart", "logr"}
        Which propensity model to use.
    **kwargs : dict, optional
        Keyword arguments passed to propensity model.

    Returns
    -------
    scores : numpy.ndarray
        Propensity score for each point in `data`.
    """

    if method == "logr":
        model = LogisticRegression
        data = PolynomialFeatures(
            2, interaction_only=True, include_bias=False
        ).fit_transform(data)

    if method == "cart":
        model = DecisionTreeClassifier

    scores = model(**kwargs).fit(data, labels).predict_proba(data)[:, 1]

    return scores


def pmse(combined, indicator, method, **kwargs):
    """Calculate the propensity score mean-squared error (pMSE).

    Parameters
    ----------
    combined : pandas.DataFrame
        The combined set of real and synthetic data.
    indicator : numpy.ndarray
        An indicator for which data are real (0) or synthetic (1).
    method : {"cart", "logr"}
        Which propensity model to use. Must be either CART (`"cart"`) or
        logistic regression with first-order interactions (`"logr"`).
    **kwargs : dict, optional
        Keyword arguments passed to propensity model.

    Returns
    -------
    float
        Propensity score mean-squared error.

    See Also
    --------
    sklearn.linear_model.LogisticRegression
    sklearn.tree.DecisionTreeClassifier

    Notes
    -----
    Propensity scores represent probabilities of group membership. By
    modelling whether an example is synthetic or not, we can use
    propensity scores as a measure of utility.

    This returns zero if the distributions are identical, and is bounded
    above by :math:`1 - c` if they are nothing alike, where :math:`c` is
    the proportion of the data that is synthetic. This method is
    therefore good for comparing multiple synthetic datasets. However,
    as this is not a test, there is no threshold distance below which we
    can claim the distributions are statistically the same.

    This function assumes that some preprocessing has been carried out
    so that the data is ready to be passed to the classification
    function. Encoding of categorical data is performed, but, for
    example, scaling is not. Without this, erroneous results may be
    returned. The logistic regression can fail to converge if many
    variables are considered. Anecdotally, this doesn't seem to
    drastically impact the propensity scores, although this should be
    investigated formally.

    Using a CART model as a classifier is recommended in the literature
    however we also support the use of logistic regression. For further
    details, see: https://doi.org/10.1111/rssa.12358
    """

    scores = _get_propensity_scores(combined, indicator, method, **kwargs)
    ideal = indicator.mean()

    return np.mean(np.square(scores - ideal))


def _pmse_logr_statistics(combined, indicator):
    """Calculate the location and scale of pMSE in the null case where
    the real and synthetic datasets are formed from identical processes.

    Parameters
    ----------

    combined : pandas.DataFrame
        Dataframe containing the combined real and synthetic data.
    indicator : numpy.ndarray
        Indicator for whether data are real (0) or synthetic (1).

    Returns
    -------
    loc : float
        Expectation of pMSE in the null case.
    scale : float
        Standard deviation of pMSE in the null case.

    Notes
    -----
    It has been shown that the null case is distributed as a multiple of
    a :math:`\\chi^2` distribution with :math:`k-1` degrees of freedom.

    Therefore, its expectation is:

    .. math::

        E(pMSE) = \\frac{(k - 1)(1 - c)^2}{N}

    and its standard deviation is:

    .. math::

        sd(pMSE) = \\frac{c \\sqrt{2(k - 1)} (1 - c)^2}{N}

    where :math:`k` is the number of predictors used in the model,
    :math:`c` is the proportion of synthetic data, and :math:`N` is the
    total number of data points.

    Here, all features and first-order interactions are used in the
    model. Let :math:`m` be the number of features, then:

    .. math::

        k = m + \\binom{m}{2}

    Further explanation and derivation of these results can be found at:
    https://doi.org/10.1111/rssa.12358
    """

    num_rows, num_cols = combined.shape
    num_predictors = num_cols + scipy.special.binom(num_cols, 2)
    prop_synth = indicator.mean()

    loc = (num_predictors - 1) * (1 - prop_synth) ** 2 / num_rows
    scale = (
        prop_synth
        * np.sqrt(2 * (num_predictors - 1))
        * (1 - prop_synth) ** 2
        / num_rows
    )

    return loc, scale


def _pmse_cart_statistics(combined, indicator, num_perms, **kwargs):
    """Estimate the location and scale of pMSE in the null case by
    repeating pMSE calculations on permuations of the indicator column
    using a CART model.

    The set of calculations are then summarised using the mean or
    standard deviation, respectively.

    Parameters
    ----------
    combined : pandas.DataFrame
        Dataframe containing the combined real and synthetic data.
    indicator : numpy.ndarray
        Indicator for whether data are real (0) or synthetic (1).
    num_perms : int
        The number of permutations to consider.
    **kwargs : dict, optional
        Keyword arguments passed to
        `sklearn.tree.DecisionTreeClassifer`.

    Returns
    -------
    loc : float
        Estimated expectation of pMSE in the null case.
    scale : float
        Estimated standard deviation of pMSE in the null case.

    Notes
    -----
    When using a CART propensity model, the number of predictors is
    unknown and the results used in `_pmse_logr_statistics` do not
    apply.

    To circumvent this, we repeatedly calculate pMSE using permutations
    of the synthetic indicator column. This should approximate the
    results for "properly" synthesised data without knowing :math:`k` a
    priori.

    Further details of this approach are available at:
    https://doi.org/10.1111/rssa.12358

    Note that the `random_state` keyword argument is used to
    (independently) create the permutations and to fit the CART model.
    Without specifying this, the results will not be reproducible.
    """

    rng = np.random.default_rng(kwargs.get("random_state", None))

    pmses = []
    for _ in range(num_perms):

        rng.shuffle(indicator)
        pmses.append(pmse(combined, indicator, method="cart", **kwargs))

    return np.mean(pmses), np.std(pmses)


def pmse_ratio(combined, indicator, method, num_perms=None, **kwargs):
    """The propensity score mean-squared error ratio.

    This is the ratio of observed pMSE to that expected under the null
    case, i.e.

    .. math::

        ratio(pMSE) = \\frac{pMSE}{E(pMSE)}

    Parameters
    ----------
    combined : pandas.DataFrame
        Dataframe containing the combined real and synthetic data.
    indicator : numpy.ndarray
        Indicator for whether data are real (0) or synthetic (1).
    method : {"cart", "logr"}
        Which propensity model to use. Must be either CART (`"cart"`) or
        logistic regression with first-order interactions (`"logr"`).
    num_perms : int, optional
        Number of permutations to consider when estimating the null case
        statistics with a CART model.
    **kwargs : dict, optional
        Keyword arguments passed to the propensity model classifier.

    Returns
    -------
    float
        The observed-to-null pMSE ratio.

    Notes
    -----
    The interpretation of this metric makes more sense for synthetic
    data. The pMSE alone gives better utility as the value gets closer
    to zero, which is only attainable when the datasets are identical.
    However, when generating synthetic data, we do not want to produce
    identical entries. Rather, we want to achieve similarity between the
    distributions of the real and synthetic datasets.

    This ratio tends towards one when this is achieved, and increases
    otherwise.

    Note that the `random_state` keyword argument is used to
    (independently) create the permutations and to fit the model when
    using a CART model. Without specifying this, the results will not be
    reproducible.
    """

    observed = pmse(combined, indicator, method, **kwargs)

    if method == "logr":
        loc, _ = _pmse_logr_statistics(combined, indicator)
    if method == "cart":
        loc, _ = _pmse_cart_statistics(
            combined, indicator, num_perms, **kwargs
        )

    return observed / loc


def pmse_standardised(combined, indicator, method, num_perms=None, **kwargs):
    """The standardised propensity score mean-squared error.

    This takes the observed pMSE and standardises it against the null
    case, i.e.

    .. math::

        stand(pMSE) = (pMSE - E(pMSE)) / sd(pMSE)

    Parameters
    ----------
    combined : pandas.DataFrame
        Dataframe containing the combined real and synthetic data.
    indicator : numpy.ndarray
        Indicator for whether data are real (0) or synthetic (1).
    method : {"cart", "logr"}
        Which propensity model to use. Must be either CART (`"cart"`) or
        logistic regression with first-order interactions (`"logr"`).
    num_perms : int, optional
        Number of permutations to consider when estimating the null case
        statistics with a CART model.
    **kwargs : dict, optional
        Keyword arguments passed to the propensity model.

    Returns
    -------
    float
        The null-standardised pMSE.

    Notes
    -----
    The interpretation of this metric makes more sense for synthetic
    data. The pMSE alone indicates better utility as it gets closer to
    zero, which is only attainable when the datasets are identical.
    However, when generating synthetic data, we do not want to produce
    identical entries. Rather, we want to achieve similarity between the
    distributions of the real and synthetic datasets.

    This standardised value tends towards zero when this is achieved,
    and increases in magnitude otherwise.

    Note that the `random_state` keyword argument is used to
    (independently) create the permutations and to fit the model when
    using a CART model. Without specifying this, the results will not be
    reproducible.
    """

    observed = pmse(combined, indicator, method, **kwargs)

    if method == "logr":
        loc, scale = _pmse_logr_statistics(combined, indicator)
    if method == "cart":
        loc, scale = _pmse_cart_statistics(
            combined, indicator, num_perms, **kwargs
        )

    return (observed - loc) / scale


def propensity_metrics(
    real, synth, method="cart", feats=None, num_perms=20, **kwargs
):
    """Propensity score-based metrics.

    This function calculates three metrics based on the propensity score
    mean-squared error (pMSE), all of which quantify utility by
    measuring the distinguishability of the synthetic data. That is, how
    readily real and synthetic data can be identified.

    To do this, the datasets are combined and their origins tracked by a
    boolean indicator. This combined dataset is then used to fit a
    binary classification model (CART or logistic regression with
    first-order interactions) with the indicator as the target. The
    propensity score for each row is then extracted and summarised to
    give a metric.

    The returned metrics are the observed pMSE along with the pMSE ratio
    and standardised pMSE. These second two metrics are given relative
    to the null case where the real and synthetic data are produced from
    identical processes.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    method : {"cart", "logr"}, default "cart"
        Which propensity model to use. Must be either CART (`"cart"`) or
        logistic regression with first-order interactions (`"logr"`).
    feats : list of str or None, default None
        List of features in the dataset to be used in the propensity
        model. If `None` (default), all common features are used.
    num_perms : int, default 20
        Number of permutations to consider when estimating the null case
        statistics with a CART model.
    **kwargs : dict, optional
        Keyword arguments passed to the propensity model.

    Returns
    -------
    observed : float
        The observed pMSE.
    standard : float
        The null-standardised pMSE.
    ratio : float
        The observed-null pMSE ratio.

    Raises
    ------
    ValueError
        If `method` is not one of `'cart'` or `'logr'`.

    See Also
    --------
    sklearn.linear_model.LogisticRegression
    sklearn.tree.DecisionTreeClassifier
    synthgauge.metrics.propensity.pmse
    synthgauge.metrics.propensity.pmse_ratio
    synthgauge.metrics.propensity.pmse_standardised

    Notes
    -----
    For the CART model, `sklearn.tree.DecisionTreeClassifier` is used.
    Meanwhile, the logistic regression model uses
    `sklearn.linear_model.LogisticRegression`.

    Note that the `random_state` keyword argument is used to
    (independently) create the permutations and to fit the model when
    using a CART model. Without specifying this, the results will not be
    reproducible.

    Details on these metrics can be found at:
    https://doi.org/10.1111/rssa.12358
    """

    if method not in ("cart", "logr"):
        raise ValueError(
            f"Propensity method must be 'cart' or 'logr' not {method}."
        )

    feats = feats or real.columns.intersection(synth.columns)
    combined, indicator = _combine_encode_and_pop(real[feats], synth[feats])

    if method == "logr":
        loc, scale = _pmse_logr_statistics(combined, indicator)
    if method == "cart":
        loc, scale = _pmse_cart_statistics(
            combined, indicator, num_perms, **kwargs
        )

    observed = pmse(combined, indicator, method, **kwargs)
    standard = (observed - loc) / scale
    ratio = observed / loc

    PropensityResult = namedtuple(
        "PropensityResult",
        ("pmse", "pmse_standardised", "pmse_ratio"),
    )

    return PropensityResult(observed, standard, ratio)



"""Univariate utility metrics."""

import numpy as np
from scipy import stats
from scipy.spatial.distance import jensenshannon


def _get_bin_counts(real, synth, feature, bins):
    """Discretise real and synthetic features, and return the bin counts
    for each. Used by the Kullback-Leibler and Jensen-Shannon functions.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data. respectively.
    feature : str
        Feature to be processed into bin counts. Must be continuous.
    bins : int or str or None
        The binning method to use. If `int`, is the number of bins. If
        `str`, must be a method accepted by `numpy.histogram_bin_edges`.
        If `None`, the feature is assumed to be categorical and counts
        are taken for every value that appears in either dataset.

    Returns
    -------
    real_counts, synth_counts : np.ndarray
        The bin counts for the real and synthetic data, respectively.
    """

    real, synth = real[feature].values, synth[feature].values
    combined = np.concatenate((real, synth))

    if bins is None:
        values = np.unique(combined)
        real_counts = np.array([sum(real == val) for val in values])
        synth_counts = np.array([sum(synth == val) for val in values])
    else:
        edges = np.histogram_bin_edges(combined, bins=bins)
        real_counts, _ = np.histogram(real, edges)
        synth_counts, _ = np.histogram(synth, edges)

    return real_counts, synth_counts


def kullback_leibler(real, synth, feature, bins="auto", **kwargs):
    """Kullback-Leibler divergence.

    Describes how much the synthetic feature distribution varies from
    the real distribution in terms of relative entropy. The divergence
    is assymmetric and does not satisfy the triangle inequality. Thus,
    it does not describe "distance" in the mathematical sense.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature of the datasets to compare. This must be continuous.
    bins : int or str or None, default "auto"
        The binning method to use. If `int`, is the number of bins. If
        `str`, must be a method accepted by `numpy.histogram_bin_edges`.
        If `None`, the feature is assumed to be categorical and counts
        are taken for each value in either dataset.
    **kwargs : dict, optional
        Keyword arguments for `scipy.stats.entropy`.

    Returns
    -------
    float
        The computed divergence between the distributions.

    See Also
    --------
    scipy.stats.entropy

    Notes
    -----
    This is a wrapper function for `scipy.stats.entropy`. Since this
    function expects a probability vector, the data is first discretised
    into evenly-spaced bins.

    We can think of the Kullback-Leibler divergence as a measure of
    surprise we might expect seeing an example from the real data,
    relative to the distribution of the synthetic.

    The divergence is zero if the distributions are identical, and
    larger values indicate that the two discretised distributions are
    further from one another. Note that the KL divergence may be
    infinite - the most common case where this happens is when the
    synthetic data does not contain any data in at least one bin.

    An optimal 'bins' value has not been suggested.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame(
    ...     {"a": [0, 1, 2, 3], "b": ["cat", "cow", "dog", "emu"]}
    ... )
    >>> synth = pd.DataFrame(
    ...     {"a": [3, 1, 2, 0], "b": ["cat", "cat", "cow", "dog"]}
    ... )

    The first feature is replicated up to a re-ordering in the synthetic
    data, so its KL divergence is zero:

    >>> kullback_leibler(real, synth, "a")
    0.0

    However, the second feature does not include the ``"emu"`` category
    in the synthetic data, so the divergence is infinite:

    >>> kullback_leibler(real, synth, "b", bins=None)
    inf

    """

    real_counts, synth_counts = _get_bin_counts(real, synth, feature, bins)

    return stats.entropy(real_counts, synth_counts, **kwargs)


def jensen_shannon_divergence(real, synth, feature, bins="auto", **kwargs):
    """Jensen-Shannon divergence.

    Also known as the information radius, the Jensen-Shannon divergence
    describes the similarity between two probability distributions in
    terms of entropy. This divergence modifies the Kullback-Leibler
    divergence to be symmetric and finite (between 0 and 1).

    The divergence does not satisfy the triangle inequality. Thus, it
    does not describe "distance" in the mathematical sense. Taking its
    square root provides a metric known as the Jensen-Shannon distance.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature of the datasets to compare. This must be continuous.
    bins : int or str or None, default "auto"
        The binning method to use. If `int`, is the number of bins. If
        `str`, must be a method accepted by `numpy.histogram_bin_edges`.
        If `None`, the feature is assumed to be categorical and counts
        are taken for each value in either dataset.
    **kwargs : dict, optional
        Keyword arguments for `scipy.spatial.distance.jensenshannon`.

    Returns
    -------
    float
        The computed divergence between the distributions.

    See Also
    --------
    synthgauge.metrics.univariate_distance.jensen_shannon_distance
    scipy.spatial.distance.jensenshannon

    Notes
    -----
    This is a wrapper of
    `synthgauge.metrics.univariate_distance.jensen_shannon_distance`,
    which in turn wraps `scipy.spatial.distance.jensenshannon`. Since
    this function expects probability vectors, the data is first
    discretised into evenly-spaced bins.

    We can think of the Jensen-Shannon divergence as the amount of
    information, or entropy, encoded in the difference between the
    real and synthetic distributions of the feature.

    The divergence is zero if the distributions are identical, and is
    bounded above by one if they are nothing alike. This method is
    therefore good for comparing multiple synthetic datasets, or
    features within a dataset, to see which is closest to the real.
    However, as this is not a test, there is no threshold distance below
    which we can claim the distributions are statistically the same.

    An optimal 'bins' value has not been suggested.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame(
    ...     {"a": [0, 1, 3, 4], "b": ["cat", "cow", "dog", "emu"]}
    ... )
    >>> synth = pd.DataFrame(
    ...     {"a": [3, 3, 2, 0], "b": ["cat", "cat", "cow", "dog"]}
    ... )

    The second feature appears to be more similar than the first across
    datasets:

    >>> jensen_shannon_divergence(real, synth, "a")
    0.1732867951399863
    >>> jensen_shannon_divergence(real, synth, "b", bins=None)
    0.10788077716941784

    """

    return jensen_shannon_distance(real, synth, feature, bins, **kwargs) ** 2


def jensen_shannon_distance(real, synth, feature, bins="auto", **kwargs):
    """Jensen-Shannon distance.

    Describes the difference between two distributions in terms of
    entropy. Calculated as the square root of the Jensen-Shannon
    divergence, the Jensen-Shannon distance satisfies the mathematical
    definition of a metric.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature of the datasets to compare. This must be continuous.
    bins : int or str or None, default "auto"
        The binning method to use. If `int`, is the number of bins. If
        `str`, must be a method accepted by `numpy.histogram_bin_edges`.
        If `None`, the feature is assumed to be categorical and counts
        are taken for each value in either dataset.
    **kwargs : dict, optional
        Keyword arguments for `scipy.spatial.distance.jensenshannon`.

    Returns
    -------
    distance : float
        The computed distance between the distributions.

    See Also
    --------
    synthgauge.metrics.univariate_distance.jensen_shannon_divergence
    scipy.spatial.distance.jensenshannon

    Notes
    -----
    This is a wrapper for `scipy.spatial.distance.jensenshannon`. Since
    this function expects probability vectors, the data is first
    discretised into evenly-spaced bins.

    We can think of the Jensen-Shannon distance as the amount of
    information, or entropy, encoded in the difference between the
    `real` and `synth` distributions of the `feature`.

    The distance is zero if the distributions are identical, and is
    bounded above by one if they are nothing alike. This method is
    therefore good for comparing multiple synthetic datasets, or
    features within a dataset, to see which is closest to the real.
    However, as this is not a test, there is no threshold distance below
    which we can claim the distributions are statistically the same.

    An optimal 'bins' value has not been suggested.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame(
    ...     {"a": [0, 1, 3, 4], "b": ["cat", "cow", "dog", "emu"]}
    ... )
    >>> synth = pd.DataFrame(
    ...     {"a": [3, 3, 2, 0], "b": ["cat", "cat", "cow", "dog"]}
    ... )

    The second feature appears to be more similar than the first across
    datasets:

    >>> jensen_shannon_distance(real, synth, "a")
    0.41627730557884884
    >>> jensen_shannon_distance(real, synth, "b", bins=None)
    0.328452092654953

    """

    real_counts, synth_counts = _get_bin_counts(real, synth, feature, bins)

    return jensenshannon(real_counts, synth_counts, **kwargs)


def wasserstein(real, synth, feature, **kwargs):
    """The (first) Wasserstein distance.

    Also known as the "Earth Mover's" distance, this metric can be
    thought of as calculating the amount of "work" required to move from
    the distribution of the synthetic data to the distribution of the
    real data.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature of the datasets to compare. This must be continuous.
    **kwargs : dict, optional
        Keyword arguments for `scipy.stats.wasserstein_distance`.

    Returns
    -------
    float
        The computed distance between the distributions.

    See Also
    --------
    scipy.stats.wasserstein_distance

    Notes
    -----
    This is a wrapper for `scipy.stats.wasserstein_distance`.
    Computationally, we can find the Wasserstein distance by calculating
    the area between the cumulative distribution functions for the two
    distributions.

    If :math:`s` is the synthetic feature distribution, :math:`r` is the
    real feature distribution, and :math:`R` and :math:`S` are their
    respective cumulative distribution functions, then

    .. math::

        W(s, r) = \\int_{-\\infty}^{+\\infty} |S - R|

    The distance is zero if the distributions are identical and
    increases as they become less alike. This method is therefore good
    for comparing multiple synthetic datasets, or features within a
    dataset, to see which is closest to the real. However, as this is
    not a test, there is no threshold distance below which we can claim
    the distributions are statistically the same.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame(
    ...     {"a": [0, 1, 3, 4, 3, 4], "b": [3, 7, 2, 1, 7, 4]}
    ... )
    >>> synth = pd.DataFrame(
    ...     {"a": [3, 3, 2, 0, 2, 3], "b": [1, 5, 1, 1, 6, 3]}
    ... )

    The first feature appears to be more similar than the second across
    datasets:

    >>> wasserstein(real, synth, "a")
    0.6666666666666667
    >>> wasserstein(real, synth, "b")
    1.166666666666667

    """

    return stats.wasserstein_distance(real[feature], synth[feature], **kwargs)


def kolmogorov_smirnov(real, synth, feature, **kwargs):
    """Kolmogorov-Smirnov test.

    The Kolmogorov-Smirnov test statistic is the maximum difference
    between the cumulative distribution functions of the real and
    synthetic features.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Name of the feature to compare. This must be continuous.
    **kwargs : dict, optional
        Keyword arguments for `scipy.stats.ks_2samp`.

    Returns
    -------
    statistic, pvalue : float
        Kolmogorov-Smirnov test statistic.
    pvalue : float
        Two-tailed p-value.

    See Also
    --------
    scipy.stats.ks_2samp

    Notes
    -----
    This is a wrapper for `scipy.stats.ks_2samp`, which tests whether
    two samples are drawn from the same distribution by calculating the
    maximum difference between their cumulative distribution functions.

    If the returned statistic is small or the p-value is high, then we
    cannot reject the hypothesis that the distributions are the same.

    This approach is only defined if the feature is continuous. The
    documentation further suggests this method works best when one of
    the samples has a size of only a few thousand.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame(
    ...     {"a": [0, 1, 3, 3, 4, 4], "b": [1, 4, 5, 6, 7, 7]}
    ... )
    >>> synth = pd.DataFrame(
    ...     {"a": [0, 2, 2, 3, 3, 3], "b": [1, 1, 1, 2, 2, 3]}
    ... )

    The first feature appears to come from the same distribution in both
    datasets.

    >>> kolmogorov_smirnov(real, synth, "a")
    KstestResult(statistic=0.3333333333333333, pvalue=0.9307359307359307)

    The second feature appears to come from different distributions in
    the datasets.

    >>> kolmogorov_smirnov(real, synth, "b")
    KstestResult(statistic=0.8333333333333334, pvalue=0.025974025974025972)

    """

    return stats.ks_2samp(real[feature], synth[feature], **kwargs)


def kruskal_wallis(real, synth, feature, **kwargs):
    """Kruskal-Wallis H test.

    The Kruskal-Wallis test seeks to determine whether two sets of data
    originated from the same distribution. This is acheived by pooling
    and ranking the datasets. A low p-value suggests the two sets
    originate from different distributions and are not similar.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature of the datasets to compare. This must be continuous.
    **kwargs : dict, optional
        Keyword arguments for `scipy.stats.kruskal`.

    Returns
    -------
    statistic : float
        The Kruskal-Wallis H statistic.
    pvalue : float
        The p-value for the test.

    See Also
    --------
    scipy.stats.kruskal

    Notes
    -----
    This is a wrapper function for `scipy.stats.kruskal`.

    The null hypothesis for this test is that the medians of the
    distributions are equal. The alternative hypothesis is then that
    they are different. This would suggest that the synthetic and real
    data are not similarly distributed.

    We notice, however, that failure to reject the null hypothesis only
    suggests that the medians could be equal and says nothing else about
    how the data are distributed.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame(
    ...     {"a": [0, 1, 3, 3, 4, 4], "b": [1, 4, 5, 6, 7, 7]}
    ... )
    >>> synth = pd.DataFrame(
    ...     {"a": [0, 2, 2, 3, 3, 3], "b": [1, 1, 1, 2, 2, 3]}
    ... )

    The test for the first feature suggests that the data are similarly
    distributed according to their medians.

    >>> kruskal_wallis(real, synth, "a")
    KruskalResult(statistic=0.5646387832699667, pvalue=0.45239722100817814)

    The second feature test is much clearer that the data are drawn from
    distributions with different medians.

    >>> kruskal_wallis(real, synth, "b")
    KruskalResult(statistic=4.877737226277376, pvalue=0.02720526089960062)

    """

    return stats.kruskal(real[feature], synth[feature], **kwargs)


def mann_whitney(real, synth, feature, **kwargs):
    """Mann-Whitney U test.

    The Mann-Whitney test compares two sets of data by examining how
    well-mixed they are when pooled. This is acheived by ranking the
    pooled data. A low p-value suggests the data are not similar.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature of the datasets to compare. This must be continuous.
    **kwargs : dict, optional
        Keyword arguments for `scipy.stats.mannwhitneyu`.

    Returns
    -------
    statistic : float
        The Mann-Whitney U statistic, in particular U for `synth`.
    pvalue : float
        Two-sided p-value assuming an asymptotic normal distribution.

    See Also
    --------
    scipy.stats.mannwhitneyu

    Notes
    -----
    This is a wrapper function for `scipy.stats.mannwhitneyu`.

    The null hypothesis for this test is that for randomly selected real
    and synthetic values, the probability that the real value is greater
    than the synthetic is the same as the probability that the synthetic
    value is greater than the real.

    We reject this hypothesis if the p-value is suitably small. This
    would in turn suggest that the synthetic and real data are not
    similarly distributed.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame(
    ...     {"a": [0, 1, 3, 3, 4, 4], "b": [1, 4, 5, 6, 7, 7]}
    ... )
    >>> synth = pd.DataFrame(
    ...     {"a": [0, 2, 2, 3, 3, 3], "b": [1, 1, 1, 2, 2, 3]}
    ... )

    If we were to choose our p-value threshold as 0.05, we would reach
    the conclusion that the distributions of the first feature are
    similar but the distributions of the second feature are not.

    >>> mann_whitney(real, synth, "a")
    MannwhitneyuResult(statistic=22.5, pvalue=0.5041764308016705)
    >>> mann_whitney(real, synth, "b")
    MannwhitneyuResult(statistic=31.5, pvalue=0.033439907088311766)

    """

    return stats.mannwhitneyu(
        real[feature], synth[feature], alternative="two-sided", **kwargs
    )


def wilcoxon(real, synth, feature, **kwargs):
    """Wilcoxon signed-rank test.

    In this use, the Wilcoxon test compares the distributions of paired
    data. It does this by ranking the pairwise differences between the
    real and synthetic data.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature of the datasets to compare. This must be continuous.
    **kwargs : dict, optional
        Keyword arguments for `scipy.stats.wilcoxon`.

    Returns
    -------
    statistic : float
        The sum of the ranks of the differences above or below zero,
        whichever is greater.
    pvalue : float
        Two-sided p-value.

    See Also
    --------
    scipy.stats.wilcoxon

    Notes
    -----
    This is a wrapper function for `scipy.stats.wilcoxon`.

    The null hypothesis for this test is that the median of the paired
    differences is zero. The alternative hypothesis is that it is
    different from zero. This would suggest that the synthetic and real
    data are not similarly distributed.

    This test only makes sense when the synthetic and real data are
    paired. That is, each synthetic datum is matched to a real one. In
    which case, it is required that data are ordered to reflect this.

    Examples
    --------
    >>> import pandas as pd
    >>> real = pd.DataFrame({"a": [0, 1, 2, 3, 4, 5, 8, 9]})
    >>> synth = pd.DataFrame({"a": [2, 0, 3, 7, 7, 4, 2, 1]})

    By eye, you might think these distributions are quite different from
    one another. However, the Wilcoxon test suggests that these two
    datasets were drawn from similar distributions.

    >>> wilcoxon(real, synth, "a")
    WilcoxonResult(statistic=17.0, pvalue=0.9453125)

    """

    return stats.wilcoxon(real[feature], synth[feature], **kwargs)



"""Utility metrics using `scikit-learn`-style classifiers."""

import inspect
from collections import namedtuple

import numpy as np
from sklearn.compose import ColumnTransformer
from sklearn.metrics import f1_score, precision_score, recall_score
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler


def _make_preprocessor(data, feats):
    """Make a pre-processing pipe for transforming numeric and
    categorical columns.

    Parameters
    ----------
    data : pandas.DataFrame
        The dataset containing at the least the columns in `feats`.
    feats : list of str
        A list of columns in `data` to be separated by data type.

    Returns
    -------
    preprocessor : sklearn.pipeline.Pipeline
        The pre-processing pipeline.
    """

    numeric = data[feats].select_dtypes(include="number").columns
    categorical = data[feats].select_dtypes(exclude="number").columns

    numeric_transformer = Pipeline(steps=[("scaler", StandardScaler())])
    categorical_transformer = Pipeline(steps=[("encoder", OneHotEncoder())])

    preprocessor = ColumnTransformer(
        transformers=[
            ("numeric", numeric_transformer, numeric),
            ("categorical", categorical_transformer, categorical),
        ]
    )

    return preprocessor


def _make_pipeline(classifier, seed, preprocessor, **kwargs):
    """Create the pipeline of data pre-processing and classification.

    Parameters
    ----------
    classifier : scikit-learn estimator
        The `scikit-learn`-style class to be used as the classifier.
    seed : int
        Random seed to use for reproducibility. Only used if
        `random_state` is a parameter of `classifier`.
    preprocessor : sklearn.pipeline.Pipeline
        The pre-processing pipeline.
    **kwargs : dict, optinal
        featsword arguments for `classifier`.

    Returns
    -------
    pipeline : sklearn.pipeline.Pipeline
        A complete classification pipeline.
    """

    classifier_params = inspect.signature(classifier).parameters
    if classifier_params.get("random_state", None):
        kwargs["random_state"] = seed

    pipeline = Pipeline(
        steps=[
            ("preprocessor", preprocessor),
            ("classifier", classifier(**kwargs)),
        ]
    )

    return pipeline


def _get_scores(test, pred):
    """Calculate the precision, recall and f1 scores for a set of
    predicted values.

    Parameters
    ----------
    test : array_like
        Labels from the test set.
    pred : array_like
        Predicted labels.

    Returns
    -------
    scores : list
        The precision, recall and f1 score given the test set and
        predicted labels.
    """

    scores = [
        score_func(test, pred, average="macro")
        for score_func in (precision_score, recall_score, f1_score)
    ]

    return scores


def classification_comparison(
    real,
    synth,
    feats,
    target,
    classifier,
    test_prop=0.2,
    random_state=None,
    **kwargs
):
    """Classification utility metric.

    This metric fits two (identical) classification models to `real` and
    `synth`, and then tests them both against withheld `real` data. We
    obtain utility scores by subtracting the precision, recall and f1
    scores of the "synthetic" model predictions from the "real" model's.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feats : list of str
        List of column names to use as the input in the classification.
    target : str
        Column to use as target in the classification.
    classifier : scikit-learn estimator
        Classifier class with `fit` and `predict` methods.
    test_prop : float or int, default 0.2
        If `float`, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the test split. If
        `int`, represents the absolute number of test samples.
    random_state : int, optional
        Random seed for shuffling during the train-test split, and for
        the classification algorithm itself.
    **kwargs : dict, optional
        featsword arguments passed to the classifier.

    Returns
    -------
    precision_difference : float
        Precision of the real model subtracted by that of the
        synthetic model.
    recall_difference : float
        Recall of the real model subtracted by that of the synthetic
        model.
    f1_difference : float
        f1 score of the real model subtracted by that of the
        synthetic model.

    Notes
    -----
    Some preprocessing is carried out before the models are trained.
    Numeric features are scaled and categorical features are
    one-hot-encoded.

    A score of zero tells us the synthetic data is just as good as the
    real at training the given classification model. Increases in these
    scores indicate poorer utility.
    """

    real_X_train, real_X_test, real_y_train, real_y_test = train_test_split(
        real[feats],
        real[target],
        test_size=test_prop,
        random_state=random_state,
    )

    synth_X_train, _, synth_y_train, _ = train_test_split(
        synth[feats],
        synth[target],
        test_size=test_prop,
        random_state=random_state,
    )

    # preprocessing
    preprocessor = _make_preprocessor(real, feats)

    # train real model, test on real
    real_pipeline = _make_pipeline(
        classifier, random_state, preprocessor, **kwargs
    ).fit(real_X_train, real_y_train.values.ravel())

    y_real_predicts_real = real_pipeline.predict(real_X_test)

    # train synth model, test on real
    synth_pipeline = _make_pipeline(
        classifier, random_state, preprocessor, **kwargs
    ).fit(synth_X_train, synth_y_train.values.ravel())

    y_synth_predicts_real = synth_pipeline.predict(real_X_test)

    # compare results
    real_scores = _get_scores(real_y_test, y_real_predicts_real)
    synth_scores = _get_scores(real_y_test, y_synth_predicts_real)
    score_differences = np.subtract(real_scores, synth_scores)

    ClassificationResult = namedtuple(
        "ClassificationResult",
        ("precision_difference", "recall_difference", "f1_difference"),
    )

    return ClassificationResult(*score_differences)



"""Mean absolute difference in feature densities."""

import numpy as np

from .. import utils


def _feature_density_diff(real, synth, feature, bins=10):
    """Computes the difference between real and synth feature densities.

    For the specified feature the density is computed across `bins` in
    both the real and synthetic data. The per-bin difference is computed
    and returned along with the bin edges that were used.

    Prior to calculating the densities. all values are converted to
    numeric via `synthgauge.utils.cat_encode`.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        The feature that will be used to compute the density.
    bins : str or int, default 10
        Bins to use for computing the density. This value is passed
        to `numpy.histogram_bin_edges` so can be any value accepted by
        that function. Default uses 10 bins.

    Returns
    -------
    hist_diff : numpy.ndarray
        The difference in feature density for each of the bins.
    bin_edges : numpy.ndarray
        The edges of the bins.
    """

    combined = utils.df_combine(real, synth, feats=[feature])
    encoded, _ = utils.cat_encode(combined, feats=[feature], return_all=True)
    enc_real, enc_synth = utils.df_separate(encoded, "source")

    bin_edges = np.histogram_bin_edges(encoded[feature], bins=bins)

    real_hist, _ = np.histogram(
        enc_real[feature], bins=bin_edges, density=True
    )
    synth_hist, _ = np.histogram(
        enc_synth[feature], bins=bin_edges, density=True
    )

    hist_diff = synth_hist - real_hist

    return hist_diff, bin_edges


def feature_density_mad(real, synth, feats=None, bins=10):
    """Mean absolute difference of feature densities.

    For each feature the difference between the density across the bins
    within `real` and `synth` is calculated. Finally the MAE across all
    features and bins is calculated. A value close to 0 indicates that
    the real and synthetic datasets have a similar set of feature
    distributions.

    Parameters
    ----------
    real : pandas.DataFrame
        DataFrame containing the real data.
    synth : pandas.DataFrame
        DataFrame containing the sythetic data.
    feats : list of str or None, default None
        The features that will be used to compute the densities. If
        `None` (default), all common features are used.
    bins : str or int, default 10
        Binning method for discretising the data. Can be anything
        accepted by `numpy.histogram_bin_edges`. Default uses 10 bins.

    Returns
    -------
    float
        Mean absolute error of feature densities.
    """

    feats = feats or real.columns.intersection(synth.columns)
    diffs = [
        _feature_density_diff(real, synth, feat, bins)[0] for feat in feats
    ]

    return np.mean(np.abs(np.concatenate(diffs)))



"""Utility metrics derived from centroid-based clustering."""

import numpy as np
import pandas as pd
from kmodes.kprototypes import KPrototypes
from sklearn.cluster import KMeans

from ..utils import df_combine


def _get_cluster_labels(combined, method, k, random_state):
    """Apply the chosen clustering method to a dataset and return its
    final labels.

    Parameters
    ----------
    combined : pandas.DataFrame
        Dataframe containing the real and synthetic data.
    method : {"kmeans", "kprototypes"}
        Which clustering method to use.
    k : int
        Number of clusters to fit.
    random_state : int, optional
        Random seed for fitting clusters.

    Returns
    -------
    labels : np.ndarray
        Integer labels indicating cluster membership for each point in
        `combined`.
    """

    if method == "kmeans":
        numeric = combined.select_dtypes(include="number")
        model = KMeans(
            n_clusters=k, algorithm="elkan", random_state=random_state
        ).fit(numeric)

    elif method == "kprototypes":
        categorical_idxs = [
            i
            for i, dtype in enumerate(combined.dtypes)
            if not pd.api.types.is_numeric_dtype(dtype)
        ]
        model = KPrototypes(
            n_clusters=k, random_state=random_state, n_init=1
        ).fit(combined, categorical=categorical_idxs)

    else:
        raise ValueError(
            'Clustering method must be one of `"kmeans"` or'
            f'`"kprototypes"` not {method}.'
        )

    return model.labels_.astype(int)


def _get_cluster_proportions(labels, indicator):
    """Calculate the proportion of each cluster that is synthetic.

    Parameters
    ----------
    labels : array_like
        Cluster membership array.
    indicator : array_like
        Indicator of which data are real and which are synthetic.

    Returns
    -------
    proportions : numpy.ndarray
        Array with synthetic data proportion of each cluster.
    """

    proportions = []
    for cluster in np.unique(labels):
        proportions.append(
            sum(indicator[labels == cluster]) / sum(labels == cluster)
        )

    return np.array(proportions)


def clustered_msd(
    real,
    synth,
    feats=None,
    method="kmeans",
    k_min=2,
    k_max=10,
    random_state=None,
):
    """(Multiple) clustered mean-squared difference (MSD).

    This metric clusters the real and synthetic data together, measuring
    the synthetic utility according to its representation across the
    fitted clusters. Since there is often no obvious choice for the
    number of clusters, :math:`k`, we consider a range of values.

    For each value of :math:`k`, the chosen clustering method is fit
    and the proportion of synthetic data in each cluster is recorded.
    The clustered MSD is then calculated as the mean-squared difference
    between these proportions and the overall proportion of synthetic
    data.

    This collection of MSDs is summarised by taking its minimum to give
    the metric value.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feats : list of str or None, default None
        Features to use in the clustering. If `None` (default), all
        common features are used.
    method : {"kmeans", "kprototypes"}, default "kmeans"
        Clustering method to use. Only k-means and k-prototypes
        are implemented. If using k-means (default), only numeric
        columns are considered, while k-prototypes allows for mixed-type
        clustering.
    k_min : int, default 2
        Minimum number of clusters to use. Default of 2.
    k_max : int, default 10
        Maximum number of clusters to use. Default of 10.
    random_state : int, optional
        The random seed used to fit the clustering algorithm.

    Returns
    -------
    float
        The minimum observed clustered MSD.

    Notes
    -----
    This function can be used with a single value of `k` by setting
    `k_min` and `k_max` both to `k`. For instance, if a sensible number
    of clusters is known a priori.

    This metric says nothing about how appropriate the clustering method
    may be for the data at hand, nor how the data are distributed among
    the clusters. Both methods considered here have rather strong
    assumptions about the relative size and characteristics of the
    clusters in the data. As such, exploratory analysis is advised to
    determine whether such centroid-based clustering is well-suited.
    """

    combined = df_combine(real, synth, feats, "source", 0, 1)
    indicator = combined.pop("source")

    msds = []
    for k in range(k_min, k_max + 1):
        labels = _get_cluster_labels(combined, method, k, random_state)
        proportions = _get_cluster_proportions(labels, indicator)
        msd = np.mean(np.square(proportions - indicator.mean()))
        msds.append(msd)

    return min(msds)



"""Functions for visually evaluating synthetic data."""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from pandas.core.dtypes.common import is_numeric_dtype

from .metrics.correlation import _pairwise_cramers_v
from .metrics.density import _feature_density_diff
from .utils import cat_encode

sns.set_theme()


def plot_histograms(df, feats=None, groupby=None, figcols=2, figsize=None):
    """Plot feature distributions.

    Plot a histogram (or countplot for categorical data) for each
    feature. Where multiple features are provided a grid will be
    generated to store all the plots.

    Optionally, a groupby feature can be specified to apply a grouping
    prior to calculating the distribution.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe containing the feature(s) to be plotted.
    feats : list of str or None, default None
        Features in to plot. If `None` (default), all features are used.
    groupby : str, optional
        Feature on which to group data.
    figcols : int, default 2
        Number of columns to use in the figure.
    figsize : tuple of float, optional
        Size of figure in inches `(width, height)`. Defaults to
        `matplotlib.pyplot.rcParams["figure.figsize"]`.

    Returns
    -------
    matplotlib.figure.Figure
    """

    feats = feats or df.columns

    n_rows = int(np.ceil(len(feats) / figcols))
    fig, axes = plt.subplots(n_rows, figcols, figsize=figsize)

    for feat, ax in zip(feats, axes.ravel()):
        plotter = sns.histplot if is_numeric_dtype(df[feat]) else sns.countplot
        plotter(data=df, x=feat, ax=ax, hue=groupby)

    # Turn off axes with no data
    for ax in axes.ravel():
        if not ax.has_data():
            ax.set_visible(False)

    fig.tight_layout()

    return fig


def _order_categorical(feat):
    """Order a feature only if it is categorical.

    Parameters
    ----------
    feat : pd.Series
        The feature to be ordered.

    Returns
    -------
    pd.Series
        The ordered feature.
    """

    return feat.cat.as_ordered() if hasattr(feat, "cat") else feat


def plot_joint(
    df, x, y, groupby=None, x_bins="auto", y_bins="auto", figsize=6
):
    """Plot bivariate and univariate graphs.

    Convenience function that leverages `seaborn`. For more granular
    control, refer to `seaborn.JointGrid` and `seaborn.jointplot`.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing the feature(s) to plot.
    x : str
        Feature to plot on the x-axis and -margin.
    y : str
        Feature to plot on the y-axis and -margin.
    groupby : str, optional
        Feature on which to group data.
    x_bins, y_bins : array_like or int or str, default "auto"
        Binning method for axis. If `array_like`, must be sequence of
        bin edges. If `int`, specifies the number of bins to use. If
        `str`, can be anything accepted by `numpy.histogram_bin_edges`.
        Defaults to `"auto"`.
    figsize: int, default 6
        Size of each side of the figure in inches (it will be square).
        Defaults to six inches.

    Returns
    -------
    seaborn.axisgrid.JointGrid
    """

    grid = sns.JointGrid(height=figsize)

    sns.histplot(
        data=df,
        x=_order_categorical(df[x]),
        y=_order_categorical(df[y]),
        hue=groupby,
        alpha=0.5,
        ax=grid.ax_joint,
    )

    # For margins can use countplot or hist depending on data type.
    # No legends are shown for these marginal plots.
    for side, feat, bins in zip(("x", "y"), (x, y), (x_bins, y_bins)):

        plot_kwargs = {
            side: feat,
            "ax": getattr(grid, f"ax_marg_{side}"),
            "hue": groupby,
        }

        if is_numeric_dtype(df[feat]):
            plotter = sns.histplot
            plot_kwargs["bins"] = bins
        else:
            plotter = sns.countplot

        ax = plotter(data=df, **plot_kwargs)
        legend = ax.get_legend()
        if legend is not None:
            legend.remove()

    return grid


def plot_histogram3d(df, x, y, x_bins="auto", y_bins="auto", figsize=None):
    """Plot 3D histogram of two features.

    This is similar to a 2D histogram plot with an extra axis added
    to display the count for each feature-wise pair as 3D bars.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing the feature(s) to plot.
    x : str
        Feature to plot on the x-axis and -margin.
    y : str
        Feature to plot on the y-axis and -margin.
    x_bins, y_bins : array_like or int or str, default "auto"
        Binning method for axis. If the corresponding feature is
        categorical, the bins will be set to the cardinality of that
        feature. If `array_like`, must be sequence of bin edges. If
        `int`, specifies the number of bins to use. If `str`, can be
        anything accepted by `numpy.histogram_bin_edges`. Defaults to
        `"auto"`.
    figsize: tuple of float, optional
        Size of figure in inches `(width, height)`. Defaults to
        `matplotlib.pyplot.rcParams["figure.figsize"]`.

    Returns
    -------
    matplotlib.figure.Figure
    """

    # Encode categorical data
    cat_feats = df.select_dtypes(include=("object", "category")).columns
    cat_labels = dict()

    if cat_feats.any():
        df, cat_labels = cat_encode(df, cat_feats, return_all=True)

    # Determine bins
    bins_xy = []
    for feat, bins in zip([x, y], [x_bins, y_bins]):
        bins = df[feat].nunique() if feat in cat_feats else bins
        bins_xy.append(np.histogram_bin_edges(df[feat], bins))

    # Create figure
    fig = plt.figure(figsize=figsize)
    ax = plt.axes(projection="3d")

    # Compute 2D histogram
    hist, xedges, yedges = np.histogram2d(df[x], df[y], bins=bins_xy)

    # Determine bar anchors and thickness
    xpos, ypos = np.meshgrid(xedges[:-1], yedges[:-1], indexing="ij")
    xpos = xpos.ravel()
    ypos = ypos.ravel()

    dx = (xedges[1] - xedges[0]) * 0.8
    dy = (yedges[1] - yedges[0]) * 0.8
    dz = hist.ravel()
    zpos = np.zeros_like(dz) + 0.1

    # Plot the 3D bars
    ax.bar3d(xpos, ypos, zpos, dx, dy, dz, zsort="average")

    # Set title and labels
    ax.set_xlabel(x, fontsize=16)
    ax.set_ylabel(y, fontsize=16)
    ax.set_zlabel("$count$", fontsize=16, rotation=1)
    ax.set_title("3D Histogram")

    # Set tick lables for categorical features
    if x in cat_labels.keys():
        ax.set_xticks(xedges[1:])
        ax.set_xticklabels(cat_labels[x])
    if y in cat_labels.keys():
        ax.set_yticks(yedges[1:])
        ax.set_yticklabels(cat_labels[y])

    return fig


def plot_correlation(
    *dfs,
    feats=None,
    method="pearson",
    plot_diff=False,
    figcols=2,
    figsize=None,
    **kwargs,
):
    """Plot correlation between features in a dataframe.

    For each dataframe provided a subplot is generated showing a
    correlation heatmap of the features. For numeric features, the
    method can be anything supported by `pandas.DataFrame.corr`; for
    categorical or object feature types `"cramers_v"` must be specified.
    If the method does not match the data type, an error is raised.

    The `plot_diff` parameter will also include a difference plot, i.e.
    correlation difference between two dataframes. This is only used
    when two dataframes are provided.

    Parameters
    ----------
    *dfs : pandas.DataFrame
        Any number of dataframes to plot.
    feats : list of str or None, default None
        Features to plot. Must be present in all elements of `dfs`.
        If `None` (default), uses features common to all dataframes.
    method : {"pearson", "spearman", "cramers_v"}, default "pearson"
        Correlation method. See `pandas.DataFrame.corr` for more details
        on `"pearson"` and `"spearman"`. When `"cramers_v"` is
        specified, correlation is calculated using
        `synthgauge.metrics.correlation.cramers_v`.
    plot_diff : bool, default False
        If `True` and exactly two dataframes are provided, will also
        plot a heatmap of the absolute differences between the
        respective datasets' correlations.
    figcols : int, default 2
        Number of columns to use in the figure. Only used when `feats`
        contains more than one feature.
    figsize : tuple of float, optional
        Size of figure in inches `(width, height)`. Defaults to
        `matplotlib.pyplot.rcParams["figure.figsize"]`.
    **kwargs : dict, optional
        Any other keyword arguments to be passed to `seaborn.heatmap`.
        For example `annot=True` will turn on cell annotations. See
        documentation for more examples.

    Raises
    ------
    ValueError
        If `method` does not match the data type(s) of `feats`.

    Returns
    -------
    matplotlib.figure.Figure
    """

    feats = feats or list(set.intersection(*(set(df.columns) for df in dfs)))

    corr_results = []
    for df in dfs:

        if method.lower() in ("pearson", "spearman"):
            data = df[feats].select_dtypes(include="number")
            if len(data.columns) == 0:
                raise ValueError(
                    f"No numeric columns available for method: {method}"
                )

            corr_results.append(
                data.corr(method=method)
                .dropna(axis=0, how="all")
                .dropna(axis=1, how="all")
            )

        if method.lower() == "cramers_v":
            data = df[feats].select_dtypes(include=["object", "category"])
            if len(data.columns) == 0:
                raise ValueError(
                    f"No categorical columns available for method: {method}"
                )

            corr_results.append(
                _pairwise_cramers_v(data)
                .dropna(axis=0, how="all")
                .dropna(axis=1, how="all")
            )

    # Get min and max to set consistant colourbar
    corr_values = np.array(corr_results)
    vmin = corr_values.min()
    vmax = corr_values.max()

    # For now only perform diff if 2 DataFrames are given; no more.
    # TODO: Allow all diff permutations?
    if len(corr_results) == 2 and plot_diff:
        corr_diff = np.abs(corr_results[0] - corr_results[1])
        corr_results.append(corr_diff)

    n_subplots = len(corr_results)
    ncols = 1 if n_subplots == 1 else figcols
    nrows = int(np.ceil(n_subplots / ncols))

    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)

    for ax_num, (ax, corr) in enumerate(
        zip(np.array(axes).ravel(), corr_results)
    ):
        if ax_num == n_subplots - 1 and plot_diff:
            sp_title = "Correlation Difference"
            # Ignore vmin and vmax for this plot as scale will be
            # different to the others
            vmin = None
            vmax = None
        else:
            sp_title = f"DataFrame {ax_num + 1} Correlation"

        sns.heatmap(corr, ax=ax, vmin=vmin, vmax=vmax, **kwargs)

        ax.set_title(sp_title)

    # Turn off axes with no data
    for ax in np.array(axes).ravel():
        if not ax.has_data():
            ax.set_visible(False)

    fig.tight_layout()
    return fig


def plot_crosstab(
    real,
    synth,
    x,
    y,
    x_bins="auto",
    y_bins="auto",
    figsize=None,
    cmap="rocket",
    **kwargs,
):
    """Plot cross-tabulation heatmap for two features.

    The two-feature crosstab calculation is performed and plotted as a
    heatmap. One heatmap is shown for the real data and one for the
    synthetic. Numeric features are discretised using the `*_bins`
    arguments.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    x : str
        Feature to plot on the x-axis and -margin.
    y : str
        Feature to plot on the y-axis and -margin.
    x_bins, y_bins : array_like or int or str, default "auto"
        Binning method for axis. If `array_like`, must be sequence of
        bin edges. If `int`, specifies the number of bins to use. If
        `str`, can be anything accepted by `numpy.histogram_bin_edges`.
        Defaults to `"auto"`. Only used for numeric features.
    figsize : tuple of float, optional
        Size of figure in inches `(width, height)`. Defaults to
        `matplotlib.pyplot.rcParams["figure.figsize"]`.
    cmap : str, default "rocket"
        Palette name for heatmap and colour bar. See the documentation
        for `seaborn.color_palette` on available palette formats.
        Defaults to `"rocket"`.
    **kwargs : dict, optional
        Any other keyword arguments to be passed to `seaborn.heatmap`.
        For example, `annot=True` will turn on cell annotations. See
        documentation for more examples.

    Returns
    -------
    matplotlib.figure.Figure
    """

    # Collect x and y values
    real_x, real_y = real[x], real[y]
    synth_x, synth_y = synth[x], synth[y]
    all_x = pd.concat((real_x, synth_x))
    all_y = pd.concat((real_y, synth_y))

    # Discretise numeric features
    if is_numeric_dtype(all_x):
        x_bins = np.histogram_bin_edges(all_x.dropna(), x_bins)
        real_x = pd.cut(real_x, x_bins)
        synth_x = pd.cut(synth_x, x_bins)

    if is_numeric_dtype(all_y):
        y_bins = np.histogram_bin_edges(all_y.dropna(), y_bins)
        real_y = pd.cut(real_y, y_bins)
        synth_y = pd.cut(synth_y, y_bins)

    freq_real = pd.crosstab(real_x, real_y, dropna=False)
    freq_synth = pd.crosstab(synth_x, synth_y, dropna=False)

    fig, axes = plt.subplots(1, 2, figsize=figsize, constrained_layout=True)

    # Use same scale for real and synth
    vmin = min(freq_real.values.min(), freq_synth.values.min())
    vmax = max(freq_real.values.max(), freq_synth.values.max())

    cmap = sns.color_palette(cmap, as_cmap=True)
    mpbl = mpl.cm.ScalarMappable(mpl.colors.Normalize(vmin, vmax), cmap=cmap)

    for freq, ax, title in zip(
        (freq_real, freq_synth), axes, ("REAL", "SYNTH")
    ):
        sns.heatmap(
            freq.T,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            cbar=False,
            ax=ax,
            **kwargs,
        )
        ax.set_title(title)

    fig.colorbar(mpbl, ax=axes, shrink=0.8)

    return fig


def plot_qq(real, synth, feature, n_quantiles=None, figsize=None):
    """Generate a Q-Q plot for a feature of real and synthetic data.

    Quantile-quantile (Q-Q) plots are used to visualise two sets of
    numeric data to see if they are generated from the same
    distribution.

    In this case, it is used to provide some insight into the
    feature distributions for the synthetic and real data. If the
    scatter plot shows a straight line, then it can be inferred that the
    two distributions are similar and therefore the synthetically
    generated data follows the same distribution as the real data.

    See `Q-Q Plot <https://en.wikipedia.org/wiki/Q-Q_plot>`_ for more
    information.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feature : str
        Feature to plot. Must be in `real` and `synth`.
    n_quantiles : int or None, default None
        Number of quantiles to calculate. If `None` (default), uses the
        length of `real`.
    figsize: tuple of float, optional
        Size of figure in inches `(width, height)`. Defaults to
        `matplotlib.pyplot.rcParams["figure.figsize"]`.

    Raises
    ------
    TypeError
        If `feature` is not a numeric data type feature.

    Returns
    -------
    matplotlib.figure.Figure
    """

    dtype = real[feature].dtype
    if not is_numeric_dtype(dtype):
        raise TypeError(f"The feature must be numeric not of type: {dtype}")

    n_quantiles = n_quantiles or len(real)

    qs = np.linspace(0, 1, n_quantiles)
    x = np.quantile(real[feature], qs)
    y = np.quantile(synth[feature], qs)

    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.scatter(x, y)
    ax.set_xlabel("real data quantiles")
    ax.set_ylabel("synth data quantiles")
    ax.set_title(f"Q-Q Plot for {feature}")

    # Plot X = Y
    min_xy = min(x.min(), y.min())
    max_xy = max(x.max(), y.max())
    ax.plot([min_xy, max_xy], [min_xy, max_xy])

    return fig


def plot_feat_density_diff(
    real, synth, feats=None, feat_bins=10, diff_bins=10, figsize=None
):
    """Plot real and synth feature density differences.

    For each feature, the density difference between `real` and `synth`
    is calculated using `metrics.density._feature_density_diff`.

    If a single feature is provided in `feats`, the plot shows the raw
    density differences for each bin in that feature.

    Where multiple features are provided, the density differences are
    pd.concatenated into a flattened array and a histogram plotted. The
    histogram represents the distribution of differences in densities
    across all features and bins.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feats : list of str or None, default None
        Features used to compute the densities. If `None` (default), all
        common features are used.
    feat_bins : str or int, default 10
        Bins to use for computing the feature densities. This value is
        passed to `numpy.histogram_bin_edges` so can be any value
        accepted by that function. By default, uses 10 bins.
    diff_bins : str or int, default 10
        Bins to use when computing the multiple-feature difference
        histogram. This value is passed to `numpy.histogram_bin_edges`
        so can be any value accepted by that function. By default, uses
        10 bins.

    Returns
    -------
    matplotlib.figure.Figure
    """

    feats = feats or real.columns.intersection(synth.columns)

    if len(feats) == 1:
        feature = feats[0]
        diff_hist, diff_edges = _feature_density_diff(
            real, synth, feature, feat_bins
        )
        xlabel = f"{feature} Binned"
        ylabel = "Density Difference"
        title = f"Feature Density Difference for {feature}"

    else:
        # TODO: option to have different bins for each feature
        diffs = [
            _feature_density_diff(real, synth, feat, feat_bins)[0]
            for feat in feats
        ]

        diff_hist, diff_edges = np.histogram(
            np.concatenate(diffs), bins=diff_bins
        )

        xlabel = "Difference Bins"
        ylabel = "Count"
        title = "Histogram of Density Differences"

    fig, ax = plt.subplots(figsize=figsize)

    # default bar width is too large so use scaled bin size
    bar_width = (diff_edges[1] - diff_edges[0]) * 0.8
    ax.bar(diff_edges[:-1], diff_hist, align="edge", width=bar_width)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

    return fig



"""Utility functions for handling real and synthetic data."""

import warnings

import pandas as pd


def df_combine(
    real,
    synth,
    feats=None,
    source_col_name="source",
    source_val_real="real",
    source_val_synth="synth",
):
    """Combine separate dataframes of real and synthetic data.

    The dataframes are concatenated along the first axis (rows) and a
    source column is added to distinguish the real data from the
    synthetic data. Optionally, specific features can be selected.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feats : list of str or None, default None
        Features to combine. If `None` (default), all common features
        are used.
    source_col_name : str, default "source"
        Name of the source column. This is added to the combined dataset
        and filled with the `source_val_real` and `source_val_synth`
        values to signify the real and synthetic data respectively.
        Defaults to `"source"`.
    source_val_real : any, default "real"
        Value to use in `source_col_name` column to signify the real
        data. Defaults to `"real"`.
    source_val_synth : any, default "synth"
        Value to use in `source_col_name` column to signify the
        synthetic data. Defaults to `"synth"`.

    Returns
    -------
    combined : pandas.DataFrame
        The combined dataframe.
    """

    feats = feats or real.columns.intersection(synth.columns)

    real = real[feats].copy()
    real[source_col_name] = source_val_real

    synth = synth[feats].copy()
    synth[source_col_name] = source_val_synth

    combined = pd.concat([real, synth], ignore_index=True)

    return combined


def df_separate(
    data,
    source_col_name,
    feats=None,
    source_val_real="real",
    source_val_synth="synth",
    drop_source_col=True,
):
    """Separate a dataframe into real and synthetic data.

    The dataframe is split using a source column and real and synthetic
    flags. Optionally, specific features can be selected.

    Parameters
    ----------
    data : pandas.DataFrame
        Dataframe to split into real and synthetic components.
    source_col_name : str
        Name of the column used to signify real versus synthetic data.
    feats : list of str or None, default None
        Features to separate. If `None` (default), uses all features.
    source_val_real : any, default "real"
        Value in `source_col_name` column signifying the real data.
        Defaults to `"real"`.
    source_val_synth : any, default "synth"
        Value in `source_col_name` column signifying the synthetic data.
        Defaults to `"synth"`.
    drop_source_col : bool, default True
        Indicates whether the `source_col_name` column should be
        dropped from the outputs (default) or not.

    Returns
    -------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    """

    feats = list(feats) if feats is not None else list(data.columns)
    feats.append(source_col_name)

    real = data[data[source_col_name] == source_val_real][feats].copy()
    synth = data[data[source_col_name] == source_val_synth][feats].copy()

    if drop_source_col:
        real.drop(columns=source_col_name, inplace=True, errors="ignore")
        synth.drop(columns=source_col_name, inplace=True, errors="ignore")

    return real, synth


def launder(real, synth, feats=None, suffix_real="real", suffix_synth="synth"):
    """Launder feature names and optionally filter.

    To provide clear distinction between the real and synthetic
    features, each dataframe is updated to append suffixes to the
    feature names. Optionally, specific features can be selected.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    feats : list of str or None, default None
        Features to launder. If `None` (default), all common features
        are used.
    suffix_real : str, default "real"
        Suffix to append to columns in `real`. Default is `"real"`.
    suffix_synth : str, default "synth"
        Suffix to append to columns in `synth`. Default is `"synth"`.

    Returns
    -------
    real : pandas.DataFrame
        Dataframe containing the laundered real data.
    synth : pandas.DataFrame
        Dataframe containing the laundered synthetic data.
    """

    feats = feats or real.columns.intersection(synth.columns)

    real = real[feats].copy()
    synth = synth[feats].copy()

    real.columns = [f"{c}_{suffix_real}" for c in real.columns]
    synth.columns = [f"{c}_{suffix_synth}" for c in synth.columns]

    return real, synth


def cat_encode(
    df, feats=None, return_all=False, convert_only=False, force=False
):
    """Convert object features to categories.

    Generates a new version of the input dataframe with the specified
    features categorically encoded with integer labels. Optionally, the
    features can be returned as `category` data type with no encoding.

    Before performing the conversion, a check is made to identify any
    speficied features that are not `object`-type and thus less suited
    to categorical encoding. A warning is raised for these features and
    they will be ignored from subsequent encoding steps unless `force`
    is set to `True`.

    Parameters
    ----------
    df : pandas.DataFrame
        Input dataframe to be converted.
    feats : list of str or None, default None
        Features in `df` to convert to categorical. If `None` (default),
        all object-type columns are selected.
    return_all : bool, default False
        If `True`, all features in `df` will be returned regardless of
        whether they were converted. If `False` (default), only the
        converted features are returned.
    convert_only : bool, default False
        If `True`, the features will only be converted to the `category`
        data-type without being integer-encoded.
    force : bool, default False
        If `True`, all features in `feats` will be encoded regardless of
        their data-type.

    Warns
    -----
    UserWarning
        A warning is given if any of the features in `feats` are not of
        an `object` data type.

    Returns
    -------
    out_df : pandas.DataFrame
        Dataframe with (at least) the converted features.
    cat_dict : dict or NoneType
        A dictionary mapping each encoded feature to its categories. If
        `convert_only=True`, returns as `None`.
    """

    all_cat_cols = df.select_dtypes(include=("object", "category")).columns
    feats = pd.Index(feats) if feats is not None else all_cat_cols

    # Check for non-object type features
    non_cat_cols = feats.difference(all_cat_cols)
    if non_cat_cols.any():
        warnings.warn(
            "Selected features include non-object types: "
            f"{non_cat_cols.to_list()}."
            "\nIs this intended? If so, rerun with `force=True`. "
            "If not, they will be dropped, unless `return_all=True`, "
            "where they will pass through unchanged."
        )

    cat_dict = {} if not convert_only else None

    feats_to_encode = feats if force else feats.difference(non_cat_cols)
    out_df = df.copy() if return_all else df[feats_to_encode].copy()

    for feature in feats_to_encode:
        out_df[feature] = out_df[feature].astype("category")

        if not convert_only:
            feature_cat = out_df[feature].cat
            cat_dict[feature] = feature_cat.categories
            out_df[feature] = feature_cat.codes

    return out_df, cat_dict



"""A library for evaluating synthetic data."""

from . import datasets, metrics, plot, utils
from .evaluator import Evaluator

__version__ = "2.1.0"

__all__ = ["Evaluator", "datasets", "metrics", "plot", "utils", "__version__"]



"""Functions for creating toy datasets."""

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification


def _adjust_data_elements(data, labels, noise, nan_prop, seed):
    """Adjust the given data and put it into a dataframe.
    This function is not intended to be used directly by users.

    Parameters
    ----------
    data : numpy.ndarray
        The data array to be adjusted.
    labels : numpy.ndarray
        A set of labels for classifying the rows of `data`.
    noise : float
        The amount of noise to inject into the data. Specifically,
        this controls the `scale` parameter of a zero-centred normal
        distribution.
    nan_prop : float
        The proportion of elements to replace with missing values.
    seed : int
        A random seed used to choose missing element indices and sample
        noise.

    Returns
    -------
    pandas.DataFrame
        The adjusted, combined dataframe.
    """

    rng = np.random.default_rng(seed)

    data = np.column_stack((data, labels))

    num_cols = data.shape[1]
    num_nans = int(data.size * nan_prop)
    nan_idxs = rng.choice(data.size, num_nans, replace=False)
    nan_coords = [(idx // num_cols, idx % num_cols) for idx in nan_idxs]

    if nan_coords:
        data[tuple(np.transpose(nan_coords))] = np.nan

    data = pd.DataFrame(data + rng.normal(scale=noise, size=data.shape))

    return data


def make_blood_types_df(noise=0, nan_prop=0, seed=None):
    """Create a toy dataset about blood types and physical atrtibutes.

    This function is used to create data for the package's examples and
    its tests. Its outputs are not intended to imply or be used for any
    meaningful data analysis.

    Parameters
    ----------
    noise : float
        Standard deviation of the Gaussian noise added to the data.
        Default is zero (no noise) and must be non-negative.
    nan_prop : float, default 0
        Proportion of dataset to replace with missing values.
    seed : int, optional
        Seed used by all random samplers. Used for reproducibility.

    Returns
    -------
    data : pandas.DataFrame
        A toy "blood type" dataset.

    Notes
    -----
    The amount of noise can be tuned to crudely simulate the creation of
    synthetic data.
    """

    X, y = make_classification(
        n_samples=1000,
        n_features=5,
        n_informative=3,
        n_redundant=2,
        n_classes=4,
        weights=[0.4, 0.3, 0.2, 0.1],
        flip_y=0.1,
        random_state=seed,
    )

    df = _adjust_data_elements(X, y, noise, nan_prop, seed)

    df.columns = [
        "age",
        "height",
        "weight",
        "hair_colour",
        "eye_colour",
        "blood_type",
    ]

    df.age = np.abs(round(df.age * (52 / 9) + 44))
    df.height = np.abs(round(df.height * (52 / 9) + 175))
    df.weight = np.abs(round(df.weight * (52 / 9) + 80))
    df.hair_colour = pd.cut(
        df.hair_colour, 4, labels=["Red", "Black", "Brown", "Blonde"]
    ).cat.as_unordered()
    df.eye_colour = pd.cut(
        df.eye_colour, 3, labels=["Blue", "Brown", "Green"]
    ).cat.as_unordered()
    df.blood_type = pd.cut(
        df.blood_type, 4, labels=["O", "A", "B", "AB"]
    ).cat.as_unordered()

    return df



"""The core class for evaluating datasets."""

import pickle
import warnings
from copy import deepcopy

import pandas as pd

from . import metrics, plot, utils


class Evaluator:
    """The central class in `synthgauge`, used to hold and evaluate data
    via metrics and visualisation.

    Parameters
    ----------
    real : pandas.DataFrame
        Dataframe containing the real data.
    synth : pandas.DataFrame
        Dataframe containing the synthetic data.
    handle_nans : str, default "drop"
        Whether to drop missing values. If yes, use "drop" (default).

    Returns
    -------
    synthgauge.Evaluator
        An `Evaluator` object ready for metric and visual evaluation.
    """

    def __init__(self, real, synth, handle_nans="drop"):
        common_feats = real.columns.intersection(synth.columns)
        ignore_feats = real.columns.union(synth.columns).difference(
            common_feats
        )

        if len(ignore_feats) > 0:
            msg = (
                f"Features {', '.join(ignore_feats)} are not common to "
                "`real` and `synth` and will be ignored in further analysis."
            )

            warnings.warn(msg)

        self.feature_names = list(common_feats)

        # Metrics is private to apply some validation
        self.__metrics = dict()  # assign metrics and kwargs
        self.metric_results = dict()  # store results

        # Handle NaNs
        if handle_nans == "drop":
            real.dropna(inplace=True)
            synth.dropna(inplace=True)

        self.real_data = real
        self.synth_data = synth
        self.combined_data = utils.df_combine(self.real_data, self.synth_data)

    def describe_numeric(self):
        """Summarise numeric features.

        This function uses `pandas.DataFrame.describe` to calculate
        summary statistics for each numeric feature in `self.real_data`
        and `self.synth_data`.

        Returns
        -------
        pandas.DataFrame
            Descriptive statistics for each numeric feature.
        """

        real, synth = utils.launder(self.real_data, self.synth_data)

        return pd.concat(
            [
                real.describe(include="number"),
                synth.describe(include="number"),
            ],
            axis=1,
        ).T.sort_index()

    def describe_categorical(self):
        """Summarise categorical features.

        This function uses `pandas.DataFrame.describe` to calculate
        summary statistics for each object-type feature in
        `self.real_data` and `self.synth_data`.

        Returns
        -------
        pandas.DataFrame
            Descriptive statistics for each object-type feature.
        """

        real, synth = utils.launder(self.real_data, self.synth_data)

        return (
            pd.concat(
                [
                    real.describe(include=["category", "object"]),
                    synth.describe(include=["category", "object"]),
                ],
                axis=1,
            )
            .T.sort_index()
            .rename(columns={"top": "most_frequent"})
        )

    def add_metric(self, name, alias=None, **kwargs):
        """Add a metric to the evaluator.

        Metrics and their arguments are recorded to be run at a later
        time. This allows metric customisation but ensures that the same
        metric configuration is applied consistently, i.e. once added,
        the parameters do not require resupplying for each execution of
        the metric. Supplying a metric alias allows the same metric to
        be used multiple times with different parameters.

        Note that `self.real_data` and `self.synth_data` will be passed
        automatically to metrics that expect these arguments. They
        should not be declared in `metric_kwargs`.

        Parameters
        ----------
        name : str
            Name of the metric. Must match one of the functions in
            `synthgauge.metrics`.
        alias : str, optional
            Alias to be given to this use of the metric in the results
            table. Allows the same metric to be used multiple times with
            different parameters. If not specified, `name` is used.
        **kwargs : dict, optional
            Keyword arguments for the metric. Refer to the associated
            metric documentation for details.
        """

        try:
            getattr(metrics, name)
            kwargs["name"] = name
            alias = name if alias is None else alias
            self.__metrics.update({alias: kwargs})

        except AttributeError:
            raise NotImplementedError(f"Metric '{name}' is not implemented")

    def add_custom_metric(self, alias, func, **kwargs):
        """Add a custom metric to the evaluator.

        A custom metric uses any user-defined function that accepts the
        real and synthetic dataframes as the first and second positional
        arguments, respectively. Any other parameters must be defined as
        keyword arguments. The metric function can return a value of any
        desired type although scalar numeric values are recommended, or
        `collections.namedtuples` when there are multiple outputs.

        Parameters
        ----------
        alias : str
            Alias for the metric to appear in the results table.
        func : function
            Top-level metric function to be called during the evaluation
            step. The first two arguments of `func` must be `self.real`
            and `self.synth`.
        **kwargs : dict, optional
            Keyword arguments to be passed to `func`.
        """

        kwargs.update({"func": func, "name": alias})
        self.__metrics.update({alias: kwargs})

    def copy_metrics(self, other):
        """Copy metrics from another evaluator.

        To facilitate consistent comparisons of different synthetic
        datasets, this function copies the metrics dictionary from
        another `Evaluator` instance.

        Parameters
        ----------
        other : Evaluator
            The other evaluator from which the metrics dictionary will
            be copied.
        """

        if not isinstance(other, Evaluator):
            raise TypeError("`other` must be of class Evaluator")

        self.__metrics = deepcopy(other.metrics)

    def save_metrics(self, filename):
        """Save the current metrics dictionary to disk via `pickle`.

        Parameters
        ----------
        filename : str
            Path to pickle file to save the metrics.
        """

        with open(filename, "wb") as f:
            pickle.dump(self.metrics, f)

    def load_metrics(self, filename, overwrite=False):
        """Load metrics from disk.

        Update or overwrite the current metric dictionary from a pickle.

        Parameters
        ----------
        filename : str
            Path to metrics pickle file.
        overwrite : bool, default False
            If `True`, all current metrics will be replaced with the
            loaded metrics. Default is `False`, which will update the
            current metric dictionary with the loaded metrics.
        """

        with open(filename, "rb") as f:
            new_metrics = pickle.load(f)

        invalid_metrics = []
        for k, v in new_metrics.items():
            if getattr(metrics, v["name"], None) is None:
                invalid_metrics.append(k)

        if len(invalid_metrics) > 0:
            invalid_str = ", ".join(invalid_metrics)
            raise ValueError(f"Invalid metrics encountered in: {invalid_str}.")

        if overwrite:
            self.__metrics = new_metrics
        else:
            self.__metrics.update(new_metrics)

    @property
    def metrics(self):
        """Return __metrics."""

        return self.__metrics

    def drop_metric(self, metric):
        """Drops the named metric from the metrics dictionary.

        Parameters
        ----------
        metric : str
            Key (name or alias, if specified) of the metric to remove.
        """

        try:
            del self.__metrics[metric]
        except KeyError:
            pass

    def evaluate(self, as_df=False):
        """Compute metrics for real and synth data.

        Run through the metrics dictionary and execute each with its
        corresponding arguments. The results are returned as either a
        dictionary or dataframe.

        Results are also silently stored as a dictionary in
        `self.metric_results`.

        Parameters
        ----------
        as_df : bool, default False
            If `True`, the results will be returned as a
            `pandas.DataFrame`, otherwise a dictionary is returned.
            Default is `False`.

        Returns
        -------
        pandas.DataFrame
            If `as_df` is `True`. Each row corresponds to a metric-value
            pair. Metrics with multiple values have multiple rows.
        dict
            If `as_df` is `False`. The keys are the metric names and
            the values are the metric values (grouped). Metrics with
            multiple values are assigned to a single key.
        """

        results = dict.fromkeys(self.__metrics.keys())

        metrics_copy = deepcopy(self.__metrics)
        for metric, kwargs in metrics_copy.items():
            metric_name = kwargs.pop("name")
            if metric_name in metrics.__dict__.keys():
                metric_func = getattr(metrics, metric_name)
            else:
                metric_func = kwargs.pop("func")
            results[metric] = metric_func(
                self.real_data, self.synth_data, **kwargs
            )

        self.metric_results = dict(results)

        if as_df:
            tidy_results = {}
            for k, v in self.metric_results.items():
                try:
                    for vk, vv in v._asdict().items():
                        tidy_results[k + "-" + vk] = vv
                except AttributeError:
                    tidy_results[k] = v

            return pd.DataFrame(tidy_results, index=["value"]).T

        else:
            return results

    def plot_histograms(self, figcols=2, figsize=None):
        """Plot grid of feature distributions.

        Convenience wrapper for `synthgauge.plot.plot_histograms`. This
        function uses the combined real and synthetic data sets and
        groups by `'source'`.
        """

        return plot.plot_histograms(
            self.combined_data,
            feats=self.feature_names,
            groupby="source",
            figcols=figcols,
            figsize=figsize,
        )

    def plot_histogram3d(
        self, data, x, y, x_bins="auto", y_bins="auto", figsize=None
    ):
        """Plot 3D histogram.

        Convenience wrapper for `synthgauge.plot.plot_histogram3d`.

        Parameters
        ----------
        data: {"real", "synth", "combined"}
            Dataframe to pass to plotting function. Either `"real"` to
            pass `self.real_data`, `"synth"` to pass `self.synth_data`
            or `"combined"` to pass `self.combined_data`.
        x : str
            Column to plot along the x-axis.
        y : str
            Column to plot alont the y-axis.
        """

        return plot.plot_histogram3d(
            getattr(self, f"{data}_data"),
            x=x,
            y=y,
            x_bins=x_bins,
            y_bins=y_bins,
            figsize=figsize,
        )

    def plot_correlation(
        self, feats=None, method="pearson", figcols=2, figsize=None, **kwargs
    ):
        """Plot a grid of correlation heatmaps.

        Convenience wrapper for `synthgauge.plot.plot_correlation`. Each
        dataset (real and synthetic) has a plot, as well as one for the
        differences in their correlations.
        """

        return plot.plot_correlation(
            self.real_data,
            self.synth_data,
            feats=feats,
            method=method,
            plot_diff=True,
            figcols=figcols,
            figsize=figsize,
            **kwargs,
        )

    def plot_crosstab(self, x, y, figsize=None, **kwargs):
        """Plot pairwise cross-tabulation.

        Convenience wrapper for `synthgauge.plot.plot_crosstab`.
        Automatically sets `real` and `synth` parameters to the
        corresponding data in `self`.
        """

        return plot.plot_crosstab(
            self.real_data,
            self.synth_data,
            x=x,
            y=y,
            figsize=figsize,
            **kwargs,
        )

    def plot_qq(self, feature, n_quantiles=None, figsize=None):
        """Plot quantile-quantile plot.

        Convenience wrapper for `synthgauge.plot.plot_qq`.

        Parameters
        ----------
        feature : str
            Feature to plot.
        **kwargs : dict, optional
            Keyword arguments to pass to `synthgauge.plot.plot_qq`.
        """

        return plot.plot_qq(
            self.real_data, self.synth_data, feature, n_quantiles, figsize
        )



"""A submodule for all utility and privacy metrics."""

from .classification import classification_comparison
from .cluster import clustered_msd
from .correlation import correlation_msd, correlation_ratio_msd
from .density import feature_density_mad
from .nist import hoc, kway_marginals
from .privacy import min_nearest_neighbour, sample_overlap_score, tcap_score
from .propensity import propensity_metrics
from .univariate import (
    jensen_shannon_distance,
    jensen_shannon_divergence,
    kolmogorov_smirnov,
    kruskal_wallis,
    kullback_leibler,
    mann_whitney,
    wasserstein,
    wilcoxon,
)

__all__ = [
    "classification_comparison",
    "clustered_msd",
    "correlation_msd",
    "correlation_ratio_msd",
    "feature_density_mad",
    "hoc",
    "jensen_shannon_distance",
    "jensen_shannon_divergence",
    "kolmogorov_smirnov",
    "kruskal_wallis",
    "kullback_leibler",
    "kway_marginals",
    "mann_whitney",
    "min_nearest_neighbour",
    "propensity_metrics",
    "sample_overlap_score",
    "tcap_score",
    "wasserstein",
    "wilcoxon",
]
